@@ -5,7 +5,6 @@
 # Notes:
 # - when using -g77 then use -DUNDERSCORE_G77 to compile f2py
 #   generated extension modules (works for f2py v2.45.241_1936 and up)
-
 import os
 
 from numpy.distutils.cpuinfo import cpu
@@ -60,12 +59,12 @@ class AbsoftFCompiler(FCompiler):
         elif self.get_version() >= '9.0':
             opt = ['-shared']
         else:
-            opt = ["-K","shared"]
+            opt = ["-K", "shared"]
         return opt
 
     def library_dir_option(self, dir):
         if os.name=='nt':
-            return ['-link','/PATH:"%s"' % (dir)]
+            return ['-link', '/PATH:%s' % (dir)]
         return "-L" + dir
 
     def library_option(self, lib):
@@ -91,12 +90,14 @@ class AbsoftFCompiler(FCompiler):
 
     def get_libraries(self):
         opt = FCompiler.get_libraries(self)
-        if self.get_version() >= '10.0':
+        if self.get_version() >= '11.0':
+            opt.extend(['af90math', 'afio', 'af77math', 'amisc'])
+        elif self.get_version() >= '10.0':
             opt.extend(['af90math', 'afio', 'af77math', 'U77'])
         elif self.get_version() >= '8.0':
-            opt.extend(['f90math','fio','f77math','U77'])
+            opt.extend(['f90math', 'fio', 'f77math', 'U77'])
         else:
-            opt.extend(['fio','f90math','fmath','U77'])
+            opt.extend(['fio', 'f90math', 'fmath', 'U77'])
         if os.name =='nt':
             opt.append('COMDLG32')
         return opt
@@ -112,11 +113,11 @@ class AbsoftFCompiler(FCompiler):
 
     def get_flags_f77(self):
         opt = FCompiler.get_flags_f77(self)
-        opt.extend(['-N22','-N90','-N110'])
+        opt.extend(['-N22', '-N90', '-N110'])
         v = self.get_version()
         if os.name == 'nt':
             if v and v>='8.0':
-                opt.extend(['-f','-N15'])
+                opt.extend(['-f', '-N15'])
         else:
             opt.append('-f')
             if v:
@@ -130,8 +131,8 @@ class AbsoftFCompiler(FCompiler):
 
     def get_flags_f90(self):
         opt = FCompiler.get_flags_f90(self)
-        opt.extend(["-YCFRL=1","-YCOM_NAMES=LCS","-YCOM_PFX","-YEXT_PFX",
-                    "-YCOM_SFX=_","-YEXT_SFX=_","-YEXT_NAMES=LCS"])
+        opt.extend(["-YCFRL=1", "-YCOM_NAMES=LCS", "-YCOM_PFX", "-YEXT_PFX",
+                    "-YCOM_SFX=_", "-YEXT_SFX=_", "-YEXT_NAMES=LCS"])
         if self.get_version():
             if self.get_version()>'4.6':
                 opt.extend(["-YDEALLOC=ALL"])
@@ -139,9 +140,9 @@ class AbsoftFCompiler(FCompiler):
 
     def get_flags_fix(self):
         opt = FCompiler.get_flags_fix(self)
-        opt.extend(["-YCFRL=1","-YCOM_NAMES=LCS","-YCOM_PFX","-YEXT_PFX",
-                    "-YCOM_SFX=_","-YEXT_SFX=_","-YEXT_NAMES=LCS"])
-        opt.extend(["-f","fixed"])
+        opt.extend(["-YCFRL=1", "-YCOM_NAMES=LCS", "-YCOM_PFX", "-YEXT_PFX",
+                    "-YCOM_SFX=_", "-YEXT_SFX=_", "-YEXT_NAMES=LCS"])
+        opt.extend(["-f", "fixed"])
         return opt
 
     def get_flags_opt(self):
@@ -151,7 +152,5 @@ class AbsoftFCompiler(FCompiler):
 if __name__ == '__main__':
     from distutils import log
     log.set_verbosity(2)
-    from numpy.distutils.fcompiler import new_fcompiler
-    compiler = new_fcompiler(compiler='absoft')
-    compiler.customize()
-    print(compiler.get_version())
+    from numpy.distutils import customized_fcompiler
+    print(customized_fcompiler(compiler='absoft').get_version())
